import argparse

from charla import chat, config, ui


def mock_args():
    p = argparse.ArgumentParser()
    p.set_defaults(**config.default_settings)
    return p.parse_args([])


def test_get_content_error():
    assert chat.get_content('https://does-not-exist') == ''
    assert chat.get_content('does-not-exist.file') == ''


def test_get_content_file():
    content = chat.get_content('./tests/test_chat.py')
    assert len(content)


def test_prompt_session():
    session = chat.prompt_session(mock_args())
    assert session.message == ui.t_prompt



#!/usr/bin/env python
import argparse

from charla import chat, config
from charla.__about__ import __version__


def main(args=None) -> None:
    """Create and execute command line interface."""

    # Settings priority: cli args > user settings > default settings.
    user_settings = config.settings(config.load())

    parent = argparse.ArgumentParser(add_help=False)
    parent.add_argument('--verbose', '-v', action='store_true', help='Verbose program output.')

    parser = argparse.ArgumentParser(description='Chat with language models.')
    parser.add_argument('--model', '-m', type=str, help='Name of language model to chat with.')
    parser.add_argument('--chats-path', type=str, help='Directory to store chats.')
    parser.add_argument('--prompt-history', type=str, help='File to store prompt history.')
    parser.add_argument('--provider', type=str, help='Name of the provider to use.')
    parser.add_argument('--message-limit', type=int, help='Maximum number of messages to send to GitHub Models service.')
    parser.add_argument('--multiline', action='store_true', help='Use multiline mode.')
    parser.add_argument('--system-prompt', '-sp',
                        type=argparse.FileType(),
                        help='File that contains system prompt to use.')
    parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')
    parser.set_defaults(**user_settings, func=chat.run)

    subparsers = parser.add_subparsers(help='Sub Commands')

    parser_settings = subparsers.add_parser('settings', help='Show current settings.')
    parser_settings.add_argument('--location', action='store_true', help='Show location of settings file.')
    parser_settings.set_defaults(func=config.manage)

    argv = parser.parse_args(args)
    argv.func(argv)


if __name__ == '__main__':
    main()



from charla import config


def test_settings():
    model = 'dummy'
    settings = config.settings({'model': model})
    assert settings['model'] == model
    assert isinstance(settings['multiline'], bool)



import argparse
import json
import sys
from pathlib import Path

from platformdirs import user_cache_dir, user_config_path, user_documents_dir

NAME = 'charla'
PROVIDER_NAMES = ['ollama', 'github']

default_settings: dict = {
    'chats_path': user_documents_dir() + f'/{NAME}/chats',
    'message_limit': 20,
    'model': '',
    'multiline': False,
    'prompt_history': user_cache_dir(appname=NAME) + '/prompt-history.txt',
    'provider': PROVIDER_NAMES[0],
}
path_settings = user_config_path(NAME).joinpath('settings.json')


def load() -> dict:
    """Return settings from settings file, if it exists."""

    if path_settings.exists():
        try:
            return json.loads(path_settings.read_text())
        except json.decoder.JSONDecodeError as err:
            print(f'Settings could not be read. {err}')
    return {}


def mkdir(path: Path, **kwds):
    """Wrapper for pathlib's mkdir."""

    try:
        path.mkdir(**kwds)
    except PermissionError as err:
        sys.exit(str(err))


def settings(user_settings: dict) -> dict:
    """Return settings based on user input."""

    default_settings.update(user_settings)
    return default_settings


def manage(argv: argparse.Namespace) -> None:
    """Handler for settings subcommand."""

    if argv.location:
        print(path_settings)
    else:
        user_settings = {k: v for k, v in vars(argv).items() if k in default_settings}
        print(json.dumps(user_settings, indent=4))



import argparse
import re
from datetime import datetime
from pathlib import Path
from typing import Any

import httpx
from html2text import html2text
from prompt_toolkit import HTML, PromptSession
from prompt_toolkit import print_formatted_text as print_fmt
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.history import FileHistory
from prompt_toolkit.key_binding import KeyBindings

from charla import client, config, ui


def get_content(location: str) -> str:
    """Return content of the given source or empty string."""

    content = ''

    if location.startswith(('http://', 'https://')):
        try:
            resp = httpx.get(location, follow_redirects=True)
            content = html2text(
                resp.text, baseurl=location
            ) if resp.headers['content-type'] == 'text/html' else resp.text
        except httpx.ConnectError as err:
            print(f'Enter an existing URL.\n{err}\n')
    else:
        try:
            content = Path(location).expanduser().read_text()
        except (FileNotFoundError, PermissionError) as err:
            print(f'Enter name of an existing file.\n{err}\n')

    return content


def prompt_session(argv: argparse.Namespace) -> PromptSession:
    """Create and return a PromptSession object."""

    session: PromptSession = PromptSession(
        message=ui.t_prompt_ml if argv.multiline else ui.t_prompt,
        history=FileHistory(argv.prompt_history),
        auto_suggest=AutoSuggestFromHistory(),
        multiline=argv.multiline
    )

    print(ui.t_help)

    bindings = KeyBindings()

    @bindings.add('escape', 'm')
    def switch_multiline(_event):
        session.multiline = not session.multiline
        session.message = ui.t_prompt_ml if session.multiline else ui.t_prompt

    @bindings.add('c-o')
    def fetch(_event):
        session.message = ui.t_open
        session.completer = PathCompleter(only_directories=False, expanduser=True)

    session.key_bindings = bindings

    return session


def run(argv: argparse.Namespace) -> None:
    """Run the chat session."""

    open_location = ''  # File name or URL to be opened.

    # Prompt used to give directions to the model at the beginning of the chat.
    system_prompt = argv.system_prompt.read() if argv.system_prompt else ''

    # Determine client class and import corresponding module.
    client_cls: Any = None
    if argv.provider == 'ollama':
        from charla.client.ollama import OllamaClient
        client_cls = OllamaClient
    elif argv.provider == 'github':
        from charla.client.github import AzureClient
        client_cls = AzureClient

    # Start model API client before chat REPL in case of model errors.
    client = client_cls(argv.model, system=system_prompt, message_limit=argv.message_limit)
    client.set_info()

    # Prompt history used for auto completion.
    history = Path(argv.prompt_history)
    config.mkdir(history.parent, exist_ok=True, parents=True)

    # Location to store chat as markdown file.
    chats_path = Path(argv.chats_path)
    config.mkdir(chats_path, exist_ok=True, parents=True)
    now = datetime.strftime(datetime.now(), '%Y-%m-%d-%H-%M-%S')
    slug = re.sub(r'\W', '-', client.model)
    chats_file = chats_path / f'{now}-{slug}.md'

    # Start the chat REPL.
    session = prompt_session(argv)
    print_fmt('Chat with:', HTML(f'<ansigreen>{argv.model}</ansigreen>'), '\n')
    if system_prompt:
        print_fmt('Using system prompt:', HTML(f'<ansigreen>{argv.system_prompt.name}</ansigreen>'), '\n')

    while True:
        try:
            if not (user_input := session.prompt()):
                continue

            # Handle OPEN command input and continue to next prompt.
            if session.message == ui.t_open:
                open_location = user_input.strip()
                session.bottom_toolbar = ui.t_open_toolbar + open_location
                session.message = ui.t_prompt_ml if session.multiline else ui.t_prompt
                session.completer = None
                continue

            # Read content from location and append it to user message.
            if open_location:
                if content := get_content(open_location):
                    user_input = user_input.strip() + '\n\n' + content
                    open_location = ''
                    session.bottom_toolbar = None
                else:
                    continue

            print(f'\n{ui.t_response}\n')
            client.generate(user_input)
            print('\n')

            save(chats_file, client)

        # Exit program on CTRL-C and CTRL-D
        except (KeyboardInterrupt, EOFError):
            break

    print(f'Saving chat in: {chats_file}')
    save(chats_file, client)
    print_fmt(HTML('<b>Exiting program.</b>'))


def save(chats_file: Path, client: client.Client) -> None:
    """Save the chat as a markdown file."""

    output = f'# Chat with: {client.model}\n\n'

    for msg in client.message_history:
        if msg['role'] == 'user':
            output += f"{ui.t_prompt}{msg['text']}\n\n"
        elif msg['role'] == 'assistant':
            output += f"{ui.t_response}\n\n{msg['text']}\n\n"

    chats_file.write_text(output)



import pytest

from charla.client import Client
from charla.client.github import AzureClient
from charla.client.ollama import OllamaClient


def test_client_init_error():
    """Ensure Client class cannot be instantiated directly."""

    with pytest.raises(TypeError):
        Client(model='phi3')


def test_client_ollama_init():
    client = OllamaClient(model='phi3')
    assert hasattr(client, 'context')


def test_client_azure():
    client = AzureClient(model='gpt-4o', message_length=10)
    assert hasattr(client.client, 'close')



import os
import sys
from collections import deque
from typing import Any

from azure.ai.inference import ChatCompletionsClient
from azure.ai.inference.models import AssistantMessage, SystemMessage, UserMessage
from azure.core.credentials import AzureKeyCredential
from azure.core.exceptions import ClientAuthenticationError, HttpResponseError

from charla.client import Client


class AzureClient(Client):
    def __init__(self, model: str, system: str = '', **kwargs):
        super().__init__(model, system, **kwargs)

        # For chatting with memory.
        self.context: Any = deque([], maxlen=kwargs.get('message_limit'))

        if not (token := os.getenv('GITHUB_TOKEN')):
            sys.exit('GITHUB_TOKEN environment variable is not set or empty.')

        endpoint = 'https://models.inference.ai.azure.com'

        self.client = ChatCompletionsClient(
            endpoint=endpoint,
            credential=AzureKeyCredential(token),
            model=model
        )

        if system:
            self.context.append(SystemMessage(content=system))
            self.add_message(role='system', text=system)


    def __del__(self):
        self.client.close()


    def generate(self, prompt: str):
        self.context.append(UserMessage(content=prompt))
        self.add_message(role='user', text=prompt)

        try:
            response = self.client.complete(messages=list(self.context), stream=True)
        except (ClientAuthenticationError, HttpResponseError) as err:
            sys.exit(f'Error: {err}')

        text = ''
        for chunk in response:
            content = chunk.choices[0].delta.content
            if content:
                text += content
                print(content, end='', flush=True)

        self.context.append(AssistantMessage(content=text))
        self.add_message(role='assistant', text=text)




# UI text

t_open = 'OPEN: '
t_open_toolbar = 'Add to prompt: '
t_prompt = 'PROMPT: '
t_prompt_ml = 'PROMPT \N{LATIN SUBSCRIPT SMALL LETTER M}\N{LATIN SUBSCRIPT SMALL LETTER L}: '
t_response = 'RESPONSE:'
t_help = '''
Press CTRL-C or CTRL-D to exit chat.
Press RETURN to send prompt in single line mode.
Press ALT+M to switch between single and multi line mode.
Press ALT+RETURN to send prompt in multi line mode.
Press CTRL-O to open a file or web page and append its content to the prompt.
Press CTRL-R or CTRL-S to search prompt history.
Press ↑ and ↓ to navigate previously entered prompts.
Press → to complete an auto suggested prompt.
'''



from contextlib import suppress

from charla.cli import main


def test_cli_help(capsys):
    with suppress(SystemExit):
        main(['-h'])
    output = capsys.readouterr().out
    assert 'language model' in output



import sys
from collections.abc import Mapping

import ollama

from charla.client import Client, ModelInfo


class OllamaClient(Client):
    def __init__(self, model: str, system: str = '', **kwargs):
        super().__init__(model, system, **kwargs)

        # For chatting with memory.
        self.context: list[int] = []

        self.client = ollama.Client()


    def set_info(self):
        """Request model info from API and set model_info property."""

        # Make sure model exists or exit program.
        try:
            info = self.client.show(self.model)
        except ollama.ResponseError as err:
            sys.exit(f'Error: {err}')

        # Save model context length in meta property, that can be expanded if useful.
        arch = info['model_info']['general.architecture']
        self.model_info = ModelInfo(
            architecture=arch,
            context_length=int(info['model_info'][f'{arch}.context_length'])
        )


    def generate(self, prompt: str):
        self.add_message(role='user', text=prompt)

        response = self.client.generate(
            model=self.model, prompt=prompt, context=self.context, stream=True, system=self.system
        )

        # Make sure system message is set only once.
        self.system = ''

        text = ''
        try:
            for chunk in response:
                if not isinstance(chunk, Mapping):
                    continue
                if not chunk['done']:
                    content = chunk['response']
                    if content:
                        text += content
                        print(content, end='', flush=True)
        except ollama.ResponseError as err:
            sys.exit(f'Error: {err}')

        if isinstance(chunk, Mapping):
            self.context = chunk['context']

        self.add_message(role='assistant', text=text)



from abc import ABC, abstractmethod
from typing import NamedTuple


class ModelInfo(NamedTuple):
    architecture: str
    context_length: int


class Client(ABC):
    def __init__(self, model: str, system: str = '', **kwargs):
        self.model = model
        self.system = system
        self.message_history: list[dict] = [] # For saving chat.
        self.message_length: int | None = kwargs.get('message_length')

    @abstractmethod
    def generate(self, prompt: str):
        pass

    def set_info(self):
        pass

    def add_message(self, role, text):
        self.message_history.append({'role': role, 'text': text})
